// Copyright (c) 2006-2018 Maxim Khizhinsky
//
// Distributed under the Boost Software License, Version 1.0. (See accompanying
// file LICENSE or copy at http://www.boost.org/LICENSE_1_0.txt)

#ifndef CDSLIB_ALGO_BASE_H
#define CDSLIB_ALGO_BASE_H

#include <cds/details/defs.h>

namespace cds {

    /// Different approaches and techniques for supporting high-concurrent data structure
    namespace algo {}

} // namespace cds

#endif // #ifndef CDSLIB_ALGO_BASE_H
