// Copyright (c) 2006-2018 Maxim Khizhinsky
//
// Distributed under the Boost Software License, Version 1.0. (See accompanying
// file LICENSE or copy at http://www.boost.org/LICENSE_1_0.txt)

#ifndef CDSLIB_OS_FREE_BSD_ALLOC_ALIGNED_H
#define CDSLIB_OS_FREE_BSD_ALLOC_ALIGNED_H

#include <cds/os/posix/alloc_aligned.h>

#endif // #ifndef CDSLIB_OS_FREE_BSD_ALLOC_ALIGNED_H

