// Copyright (c) 2006-2018 Maxim Khizhinsky
//
// Distributed under the Boost Software License, Version 1.0. (See accompanying
// file LICENSE or copy at http://www.boost.org/LICENSE_1_0.txt)

#ifndef CDSLIB_GC_DEFAULT_GC_H
#define CDSLIB_GC_DEFAULT_GC_H

#include <cds/gc/hp.h>

namespace cds { namespace gc {

    /// Default garbage collector
    typedef HP     default_gc;

}} // namespace cds::gc


#endif  // #ifndef CDSLIB_GC_DEFAULT_GC_H
