// Copyright (c) 2006-2018 Maxim Khizhinsky
//
// Distributed under the Boost Software License, Version 1.0. (See accompanying
// file LICENSE or copy at http://www.boost.org/LICENSE_1_0.txt)

#ifndef CDSLIB_INTRUSIVE_SKIP_LIST_DHP_H
#define CDSLIB_INTRUSIVE_SKIP_LIST_DHP_H

#include <cds/gc/dhp.h>
#include <cds/intrusive/impl/skip_list.h>

#endif // CDSLIB_INTRUSIVE_SKIP_LIST_DHP_H
