// Copyright (c) 2006-2018 Maxim Khizhinsky
//
// Distributed under the Boost Software License, Version 1.0. (See accompanying
// file LICENSE or copy at http://www.boost.org/LICENSE_1_0.txt)

#include <cds/container/ellen_bintree_set_hp.h>
#include "test_ellen_bintree_update_desc_pool.h"

namespace cds_test {
    pool_type      s_Pool;
    lazy_pool_type s_LazyPool;
}
