// Copyright (c) 2006-2018 Maxim Khizhinsky
//
// Distributed under the Boost Software License, Version 1.0. (See accompanying
// file LICENSE or copy at http://www.boost.org/LICENSE_1_0.txt)

#ifndef CDSLIB_THREADING_DETAILS_PTHREAD_H
#define CDSLIB_THREADING_DETAILS_PTHREAD_H

#include <cds/threading/details/pthread_manager.h>

#endif // #ifndef CDSLIB_THREADING_DETAILS_PTHREAD_H
